//
// Copyright 2021 The Sigstore Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package log

type LoggerImpl interface {
	Error(...any)
	Errorf(string, ...any)
	Panic(...any)
	Info(...any)
	Infof(string, ...any)
}

var Logger LoggerImpl = &discardLogger{}

type discardLogger struct{}

func (l *discardLogger) Panic(err ...any) {
	panic(err)
}
func (l *discardLogger) Error(...any)          {}
func (l *discardLogger) Errorf(string, ...any) {}
func (l *discardLogger) Infof(string, ...any)  {}
func (l *discardLogger) Info(...any)           {}
