// Code generated by go-swagger; DO NOT EDIT.
//
// Copyright 2021 The Sigstore Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

// Package restapi Rekor
//
//	Rekor is a cryptographically secure, immutable transparency log for signed software releases.
//	Schemes:
//	  http
//	Host: rekor.sigstore.dev
//	BasePath: /
//	Version: 1.0.0
//
//	Consumes:
//	  - application/json
//
//	Produces:
//	  - application/x-pem-file
//	  - application/json
//
// swagger:meta
package restapi
